//-*-c++-*------------------------------------------------------------
//
// File name : bioArithGreater.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue May  5 11:25:34 2009
//
//--------------------------------------------------------------------

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithGreater.h"
#include "bioArithConstant.h"
#include "bioExpressionRepository.h"

#include "bioLiteralRepository.h"
bioArithGreater::bioArithGreater(bioExpressionRepository* rep,
				 patULong par,
                                 patULong left, 
                                 patULong right,
				 patError*& err) 
  : bioArithBinaryExpression(rep,par,left,right,err)
{
}

bioArithGreater::~bioArithGreater() {}


patString bioArithGreater::getOperatorName() const {
  return (">") ;
}


patReal bioArithGreater::getValue(patBoolean prepareGradient, patError*& err)  {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal left = leftChild->getValue(prepareGradient,err)  ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal right = rightChild->getValue(prepareGradient,err)  ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  return(left > right) ;

}


bioExpression* bioArithGreater::getDerivative(patULong aLiteralId, 
					      patError*& err) const {
  if (dependsOf(aLiteralId)) {
    stringstream str ;
    patString theParam = bioLiteralRepository::the()->getName(aLiteralId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    str << "Differentiation of a boolean function with respect to " << theParam << ": " << *this ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  else {
    bioExpression* result = theRepository->getZero() ;
    return result ;
  }
}


bioArithGreater* bioArithGreater::getDeepCopy(bioExpressionRepository* rep,
					      patError*& err) const {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithGreater* theNode = 
    new bioArithGreater(rep,patBadId,leftClone->getId(),rightClone->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return theNode ;
}

bioArithGreater* bioArithGreater::getShallowCopy(bioExpressionRepository* rep,
					      patError*& err) const {
  bioArithGreater* theNode = 
    new bioArithGreater(rep,patBadId,leftChild->getId(),rightChild->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return theNode ;
}


patString bioArithGreater::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patBoolean rSimple = rightChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  str << getCppResultDeclaration(prefix) << " = (" ;
  if (lSimple) {
    str << "(" << vLeft << ")" ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << ">" ;
  if (rSimple) {
    str << "(" << vRight << ")" ;    
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ")?1:0;" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patString bioArithGreater::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "(((" << vLeft << ") > (" << vRight << "))?1:0)" ;
  return patString(str.str()) ;
}





patString bioArithGreater::getExpressionString() const {
  stringstream str ;
  str << ">o" ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}

// bioArithListOfExpressions* bioArithGreater::getFunctionAndDerivatives(vector<patULong> literalIds, 
// 								   patError*& err) const {
  
//   vector<patULong> theExpressions ;
//   theExpressions.push_back(getId()) ;
//   bioArithConstant* zero = new bioArithConstant(patBadId,0.0) ;
//   for (patULong i = 1 ; i <= literalIds.size() ; ++i) {
//     if (dependsOf(literalIds[i])) {
//       err = new patErrMiscError("No derivative for absolute value") ;
//       WARNING(err->describe()) ;
//       return NULL ;
//     }
//     theExpressions.push_back(zero->getId()) ;
//   }

//   bioArithListOfExpressions* result = new bioArithListOfExpressions(NULL, theExpressions) ;
//   return result ;

// }

bioFunctionAndDerivatives* bioArithGreater::getNumericalFunctionAndGradient(vector<patULong> literalIds,patBoolean computeHessian,
							   patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size(),0.0) ;
  }
  result.theFunction = getValue(patFALSE,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return &result ;

}
