//-*-c++-*------------------------------------------------------------
//
// File name : bioArithLess.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue May  5 11:28:36 2009
//
//--------------------------------------------------------------------

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithLess.h"
#include "bioArithConstant.h"
#include "bioLiteralRepository.h"
#include "bioExpressionRepository.h"

bioArithLess::bioArithLess(bioExpressionRepository* rep,
			   patULong par,
                           patULong left, 
                           patULong right,
			   patError*& err) 
  : bioArithBinaryExpression(rep,par,left,right,err)
{
}

bioArithLess::~bioArithLess() {}

patString bioArithLess::getOperatorName() const {
  return ("<") ;
}


patReal bioArithLess::getValue(patBoolean prepareGradient, 
			       patError*& err)  {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }

  patReal left = leftChild->getValue(prepareGradient,err)  ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal right = rightChild->getValue(prepareGradient,err)  ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  return(left < right) ;
}


bioExpression* bioArithLess::getDerivative(patULong aLiteralId, 
					   patError*& err) const {
  if (dependsOf(aLiteralId)) {
    err = new patErrMiscError("No derivative for boolean functions") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  else {
    bioExpression* result = theRepository->getZero() ;
    return result ;
  }

}


bioArithLess* bioArithLess::getDeepCopy(bioExpressionRepository* rep,
					patError*& err) const {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithLess* theNode = 
    new bioArithLess(rep,patBadId,leftClone->getId(),rightClone->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}

bioArithLess* bioArithLess::getShallowCopy(bioExpressionRepository* rep,
					patError*& err) const {
  bioArithLess* theNode = 
    new bioArithLess(rep,patBadId,leftChild->getId(),rightChild->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}


patString bioArithLess::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patBoolean rSimple = rightChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  str << getCppResultDeclaration(prefix) << " = (" ;
  if (lSimple) {
    str << "(" << vLeft << ")" ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << "<" ;
  if (rSimple) {
    str << "(" << vRight << ")" ;    
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ")?1:0;" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patString bioArithLess::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "(((" << vLeft << ") < (" << vRight << "))?1:0)" ;
  return patString(str.str()) ;
}


patString bioArithLess::getExpressionString() const {
  stringstream str ;
  str << '<' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}



bioFunctionAndDerivatives* bioArithLess::getNumericalFunctionAndGradient(vector<patULong> literalIds,patBoolean computeHessian,
							   patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size(),0.0) ;
  }
  result.theFunction = getValue(patFALSE,err) ;

  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return &result ;

}
