//-*-c++-*------------------------------------------------------------
//
// File name : bioArithLog.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue May  5 14:05:01 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithLog.h"
#include "bioArithConstant.h"
#include "bioArithDivide.h"
#include "bioArithCompositeLiteral.h"
#include "bioArithNamedExpression.h"
#include "bioLiteralRepository.h"
#include "bioExpressionRepository.h"

bioArithLog::bioArithLog(bioExpressionRepository* rep,
			 patULong par,
                         patULong left,
			 patError*& err) 
  : bioArithUnaryExpression(rep, par,left,err), theLiteral(NULL), theLiteralCompId(patBadId) {

}

bioArithLog::~bioArithLog() {}

patString bioArithLog::getOperatorName() const {
  return ("log") ;
}

patReal bioArithLog::getValue(patBoolean prepareGradient, patError*& err)  {
  if (child == NULL) {
    err = new patErrNullPointer("patArithNode") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal childValue = child->getValue(prepareGradient,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (childValue == 1.0) {
    return 0.0 ;
  }
  patReal result = log(childValue) ;
  if (performCheck) {
    if (!finite(result)) {
      return -patMaxReal/2.0 ; 
    }
    else {
      return result ;
    }
  } 
  else {
    return result ;
  }
}


bioExpression* bioArithLog::getDerivative(patULong aLiteralId, patError*& err) const {
  if (child == NULL) {
    err = new patErrNullPointer("patArithNode") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  
  bioExpression* value = child->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* deriv = child->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* result = new bioArithDivide(theRepository,patBadId,deriv->getId(),value->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return result ;
}


bioArithLog* bioArithLog::getDeepCopy(bioExpressionRepository* rep,
				      patError*& err) const {
  bioExpression* leftClone(NULL) ;
  if (child != NULL) {
    leftClone = child->getDeepCopy(rep,err) ;
  }
  bioArithLog* theNode = 
    new bioArithLog(rep,patBadId,leftClone->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}


bioArithLog* bioArithLog::getShallowCopy(bioExpressionRepository* rep,
				      patError*& err) const {
  bioArithLog* theNode = 
    new bioArithLog(rep,patBadId,child->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}


patString bioArithLog::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = child->getCppCode(prefix, err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }


  stringstream str ;
  str << vLeft << endl ;
  str << getCppResultDeclaration(prefix) << " ;" << endl ;
  str << "if (!finite(log(" << child->getCppResult(prefix) <<  "))) {" << endl ;
  
  //  str << "if (" << child->getCppResult(prefix) << " <= " << patEPSILON << ") {" << endl ;
  str << "  " << getCppResult(prefix)   << "= " << -patMaxReal/2.0 << ";" << endl ;
  str << "}" << endl ;
  str << "else {" << endl ;
  str << "   " << getCppResult(prefix) << " = log(" << child->getCppResult(prefix) << ");" << endl ;
  str << "}" << endl ;
  
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
 
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patString bioArithLog::getSimpleCppCode(patError*& err) {
  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  patString vLeft = child->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "log(" << vLeft << ")" ;
  return patString(str.str()) ;
}

patBoolean bioArithLog::isSimple() const {
  return patFALSE ;
}

patString bioArithLog::getExpressionString() const {
  stringstream str ;
  str << "$L" ;
  if (child != NULL) {
    str << '{' << child->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;
}


bioFunctionAndDerivatives* bioArithLog::getNumericalFunctionAndGradient(vector<patULong> literalIds, patBoolean computeHessian,patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }
  bioFunctionAndDerivatives* c = child->getNumericalFunctionAndGradient(literalIds,computeHessian,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }

  if (c->theFunction == 0) {
    result.theFunction = patMaxReal ;
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      if (c->theGradient[i] == 0) {
	result.theGradient[i] = 0.0 ;
      }
      else {
	if (c->theGradient[i] != 0.0) {
	  result.theGradient[i] = c->theGradient[i] / c->theFunction ;
	  if (performCheck) {
	    if (!finite(result.theGradient[i])) {
	      DEBUG_MESSAGE(c->theGradient[i] << " / " << c->theFunction << " = " << result.theGradient[i]) ;
	      
	    }
	  }
	}
	else {
	  result.theGradient[i] = 0.0 ;
	}
      }
    }
  }
  else if (c->theFunction == 1) {
    result.theFunction =  0.0 ;
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      result.theGradient[i] = c->theGradient[i] ;
    }
  }
  else {
    result.theFunction =  log(c->theFunction) ;
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      if (c->theGradient[i] != 0.0) {
	result.theGradient[i] = c->theGradient[i] / c->theFunction ;
      }
      else {
	result.theGradient[i] = 0.0 ;
      }
    }
  }

  if (result.theHessian != NULL && computeHessian) {
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      for (patULong j = i ; j < literalIds.size() ; ++j) {
	patReal v = -result.theGradient[i] * c->theGradient[j] / c->theFunction ;
	//	DEBUG_MESSAGE("v["<<i<<"]["<<j<<"] = " << v) ;
	patReal hs = c->theHessian->getElement(i,j,err) ;
	//	DEBUG_MESSAGE("hs["<<i<<"]["<<j<<"] = " << v) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	if (hs != 0.0) {
	  v += hs / c->theFunction ;
	}
	//	DEBUG_MESSAGE("v["<<i<<"]["<<j<<"] = " << v) ;
	result.theHessian->setElement(i,j,v,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
      }
    }
  }

  //  DEBUG_MESSAGE("Result of log: " << result) ;
  return &result ;  
}
