//-*-c++-*------------------------------------------------------------
//
// File name : bioExpression.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Wed Apr 29 14:31:38 2009
//
//--------------------------------------------------------------------

#include <sstream>
#include "patMath.h"
#include "bioExpression.h"
#include "patDisplay.h"

#include "bioParameters.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithConstant.h"
#include "bioExpressionRepository.h"
#include "bioLiteralRepository.h"

#include "bioArithNamedExpression.h"
bioExpression::bioExpression(bioExpressionRepository* rep, patULong par) : 
  parent(par),
  codeMustBeGenerated(patTRUE), 
  __x(NULL), 
  __draws(NULL),
  theSample(NULL),
  theRepository(rep){
  

  theId = theRepository->addExpression(this) ;
  performCheck = bioParameters::the()->getValueInt("moreRobustToNumericalIssues") != 0 ;
}

ostream& operator<<(ostream& str, const bioExpression& x) {
  patError* err = NULL ;
  str << x.getExpression(err) ;
  if (err != NULL) {
    str << err->describe() << '\t' ;
  }
  return str ;
}

// Memory management for the expression nodes is not well designed. So
// memory leak will appear but it is expected that it will not be
// major.
bioExpression::~bioExpression() {
  // Free memory !!!
}

patBoolean bioExpression::isTop() const {
  return parent == patBadId;
}


void bioExpression::setParent(patULong par, patError*& err) {
  parent=par;
}
  
patULong bioExpression::getParent() const  {
  return parent;
}

patString bioExpression::getInfo() const {
  stringstream str ;
  str << "[" << getId() << "]" ;
  if (isTop()) {
    str << "[" << *this << "]"  ;
    return patString(str.str());
  }
  //  str << "[" << *this << "] within [" << *getRoot() << "]" ;
  str << "[" << *this << "]" ;
  return patString(str.str()) ;
}

patString bioExpression::getUniqueName() const {
  stringstream str ;
  str << "bio_" << theId ;
  return patString(str.str()) ;
}


patULong bioExpression::getId() const {
  return theId ;
}


patString bioExpression::getCppResult(patString prefix) const {
  stringstream str ;
  str << prefix << "_r" << theId ;
  return patString(str.str()) ;
}

patString bioExpression::getGslFunction(patString prefix, 
					patULong literalId, 
					patULong parentId,
					patError*& err)  {
  stringstream str ;
  str << " patReal  bioMyGhFunction_" << prefix << parentId << "::getValue(patReal x) {" << endl ;
  patString theLiteral = bioLiteralRepository::the()->getCppCode(literalId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  str << "double " << theLiteral << "= x ;" << endl ;
  str << "   patVariables* __x(theParams->__x) ;" << endl ;
  str << "   vector< vector<patReal> >* __draw(theParams->__draw);" << endl ;
  str << "   trVector* __beta(theParams->__beta) ;" << endl ;
  str << "   bioSample* sample(theParams->sample) ;" << endl ;
  str << "   bioIteratorSpan __theThreadSpan(theParams->__theThreadSpan) ;" << endl ;
  
  if (isSimple()) {
    str << " return " << getSimpleCppCode(err) << ";" << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    str << getCppCode(prefix,err) << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString() ;
    }
    str << " return("<< getCppResult(prefix) <<") ;" << endl ;
    
  }
  str << "}" << endl  ;
  return patString(str.str()) ;
}


patString bioExpression::getCppResultDeclaration(patString prefix) const {
  stringstream str ;
  str << "patReal " << getCppResult(prefix) ;
  return patString(str.str()) ;
}

patBoolean bioExpression::isSimple() const {
  return patTRUE ;
}

patBoolean bioExpression::isStructurallyZero() const {
  return patFALSE ;
}

patBoolean bioExpression::isStructurallyOne() const {
  return patFALSE ;
}

patString bioExpression::theIterator() const {
  return patString() ;
}

void bioExpression::setAdditionalCode(patString s) {
  additionalCode = s ;
}

patBoolean bioExpression::operator!=(const bioExpression& x) {
  return (!(*this == x)) ;
}

patBoolean bioExpression::operator==(const bioExpression& x) {
  patString me = getExpressionString() ;
  patString him = x.getExpressionString() ;
  return (me == him) ;

}



patBoolean bioExpression::isLiteral() const {
  return patFALSE ;
}


// bioArithNamedExpression* bioExpression::createNamedExpression(patError*& err) const {
  
//   if (isNamedExpression()) {
//     return (bioArithNamedExpression*)this ;
//   }

//   if (isConstant() || isLiteral()) {
//     return NULL ;
//   }

//   pair<patULong,patULong> ids = bioLiteralRepository::the()->addCompositeLiteral(err) ;
//   if (err != NULL) {
//     WARNING(err->describe()) ;
//     return NULL ;
//   }

//   bioArithNamedExpression* theNamedExpr = new bioArithNamedExpression(theRepository,patBadId,ids.second,getId(),err) ;
//   if (err != NULL) {
//     WARNING(err->describe()) ;
//     return NULL ;
//   }
  
//   return theNamedExpr ;
// }


// bioArithNamedExpression* bioExpression::createNamedExpression(patString name, patError*& err) const {
  
//   bioExpression* theCopy = getDeepCopy(err) ;
//   if (err != NULL) {
//     WARNING(err->describe()) ;
//     return NULL ;
//   }
//   bioArithNamedExpression* theNamedExpr = new bioArithNamedExpression(patBadId,name,theCopy->getId(),err) ;
//   if (err != NULL) {
//     WARNING(err->describe()) ;
//     return NULL ;
//   }
//   return theNamedExpr ;
// }



// bioArithListOfExpressions* bioExpression::
// getFunctionAndDerivatives(vector<patULong> literalIds, 
//  			  patError*& err) const {
  
//   stringstream str ;
//   str << "Not implemented for expression " << getOperatorName() ;
//   err = new patErrMiscError(str.str()) ;
//   WARNING(err->describe()) ;
//   return NULL ;
// }

//   bioArithNamedExpression* fct = createNamedExpression(err) ;
//   if (err != NULL) {
//     WARNING(err->describe()) ;
//     return NULL ;
//   }
//   vector<patULong> theExpressions ;
//   theExpressions.push_back(fct->getId()) ;
  
//   for (vector<patULong>::const_iterator i = literalIds.begin() ;
//        i != literalIds.end() ;
//        ++i) {
//     bioExpression* deriv = getDerivative(*i,err) ;
//     if (err != NULL) {
//       WARNING(err->describe()) ;
//       return NULL ;
//     }
//     bioExpression* namedDeriv = deriv->createNamedExpression(err) ;
//     if (err != NULL) {
//       WARNING(err->describe()) ;
//       return NULL ;
//     }
//     theExpressions.push_back(namedDeriv->getId()) ;
    
//   } 
//   bioArithListOfExpressions* result = new bioArithListOfExpressions(patBadId,theExpressions) ;
//   return result ;

// }

patBoolean bioExpression::isConstant() const {
  return patFALSE ;
}

patBoolean bioExpression::isSequence() const {
  return patFALSE ;
}

bioArithLikelihoodFctAndGrad* bioExpression::getLikelihoodFunctionAndDerivatives(vector<patULong> literalIds, 
										 patError*& err) const {
  err = new patErrMiscError("This function is valid only for a sum iterator.") ;
  WARNING(err->describe()) ;
  return NULL ;

}
patBoolean bioExpression::isSumIterator() const {
  return patFALSE ;
}

bioExpression* bioExpression::getParentPtr() const {
  patULong parentId = getParent() ;
  if (parentId == patBadId) {
    return NULL ;
  }
  return theRepository->getExpression(parentId) ;
}

patBoolean bioExpression::isNamedExpression() const {
  return patFALSE ;
}

trHessian* bioExpression::getBhhh() {
  return NULL ;
}



patBoolean bioExpression::isBayesian() const {
  return patFALSE ;
}


patBoolean bioExpression::isSimulator() const {
  return patFALSE ;
}

patULong bioExpression::getThreadId() const {
  if (theRepository == NULL) {
    return patBadId ;
  }
  return theRepository->getThreadId() ;
}

bioFunctionAndDerivatives* bioExpression::getNumericalFunctionAndFinDiffGradient(vector<patULong> literalIds, patError*& err)  {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }
  result.theFunction = getValue(patFALSE,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  patReal sqrteta = pow(patEPSILON, 0.5);

  patVariables beta = bioLiteralRepository::the()->getBetaValues(patFALSE) ;
  patVariables betaPlus = beta ;

  for (patULong j = 0; j < beta.size() ; ++j) {
    patReal stepsizej = sqrteta*
      patMax(patAbs(beta[j]), 1.0)*patSgn(beta[j]);
    patReal tempj =  betaPlus[j];
    betaPlus[j] += stepsizej ;
    stepsizej = betaPlus[j] - tempj;
    bioLiteralRepository::the()->setBetaValues(betaPlus,err) ; 
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    patReal fj = getValue(patFALSE,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    result.theGradient[j] = ((fj - result.theFunction)/stepsizej); 
    betaPlus[j] = tempj ;
  }
  return &result ;
  
}
